% 07/20/2022
% Nicholas Harn
% Data generation for core barrel system
clear;
close all;
clc;
rng(1);

%% System dimensions
% Number of degrees of freedom
ss_mat.n_dof    = 3;
% Number of outputs
ss_mat.m_out    = 4;
% Number of process noise inputs to the system (turbulent force)
ss_mat.n_w_in   = 3;
% Number of known inputs to the system (contact force)
ss_mat.n_u_in   = 4;

%% Core Barrel Structure
% Properties known from Western-type PWR specs:
rho     = 27678*0.285;  % Density of ASTM-304 Stainless Steel

% Properties approximated from Eastern-type PWR specs: WWER-1000 (or 400)
w_core  = 3.67;         % Outer diameter of core barrel (could be 3.37 m instead)
H       = 10.51;        % Height of core barrel (could be 10.96 m instead)
thick   = 0.065;        % Thickness of core barrel (between 0.05 and 0.08 m)

% Calculated inertia, assuming thick-walled cylinder w/ uniform mass
% distribution (i.e. h_G = H/2 = height of CoM):
m       = pi*rho*H*(w_core^2 - (w_core-2*thick)^2)/4;                       % Core mass (kg)
h_G     = H/2;                                                              % Height of center of mass (m)
Igxx    = m*( (3*(w_core/2)^2 + H^2) - (3*(w_core/2 - thick)^2 + H^2) )/12;	% Mass moment of inertia

%% System properties
% m xddot + c_damp xdot + k x = b1 w + b2 u

% Expected/guessed values (desire wn = 10 Hz for pure vertical motion)
delta_z = 0.0025;               % Assume 2.5 mm static displacement for 2 vertical hold-down springs
kz      = (m*9.81/delta_z)/2;   % Estimated spring rate of each hold-down spring
delta_y = 0.5;                  % Assume 50 cm static displacement for horizontal springs & radial keys
ky      = m*9.81/delta_y;       % Estimated spring rate of each horizontal spring/radial key (N/m)
zeta    = 0.1;                  % Expected damping ratio for horizontal support & radial key (0.05?)

% Process noise coefficients (scaled by 1000, NOT 1!!!)
by1     = 1000;
bz1     = 1000;
btheta1 = 1000;

% Mass matrix
M       = [ m,  0,  0; ...
            0,  m,  0; ...
            0,  0,  Igxx ];

% Spring matrix
K       = [ 2*ky,          0,      -2*ky*(H-h_G);   ...
            0,              2*kz,   0;              ...
            -2*ky*(H-h_G),  0       2*( ky*(H-h_G)^2 + kz*w_core^2/4 ) ];

% Locate eigenvalues (Omega2) and eigenvectors (Phi)
[Phi,Omega2]    = eig(M\K);
Omega           = sqrt(Omega2);

% Damping matrix
% C_damp          = 3.6*M + 0.001*K;
C_damp          = M*Phi*(2*zeta*Omega)/Phi;

save('c_damp.mat', 'C_damp');

% Process noise coefficient matrix
B1_coef     = [ by1,    0,       0;
                0,       bz1,     0;
                0,       0,       btheta1 ];

% Input coefficient matrix
B2_coef     = [ 1,      1,      0,          0; ...
                0,      0,      1,          1; ...
                h_G,    h_G,    -w_core/2,  w_core/2 ];

%% Create the CT SS
% Dynamics matrix
ss_mat.A    = [ zeros(ss_mat.n_dof),    eye(ss_mat.n_dof);  ...
                -M\K,                   -M\C_damp ];

% Process noise input matrix
ss_mat.B1   = [ zeros(ss_mat.n_dof, ss_mat.n_w_in);	...
                M\B1_coef ];
% Contact force input matrix
ss_mat.B2   = [ zeros(ss_mat.n_dof, ss_mat.n_u_in);	...
                M\B2_coef ];

% Measurement output matrix coefficients
coef_y      = 0.25;     % Current's dependence on y
coef_z      = 0.25;     % Current's dependence on z
coef_theta  = 0.5;      % Current's dependence on theta

% Output matrix for all ex-core sensor measurements
ss_mat.C1   = [ coef_y,    coef_z,    -coef_theta;  ...   % Sensor 1
                coef_y,    -coef_z,   coef_theta;   ...   % Sensor 2
                -coef_y,   coef_z,    coef_theta;   ...   % Sensor 3
                -coef_y,   -coef_z,   -coef_theta ];      % Sensor 4
% ss_mat.C1 = eye(ss_mat.m_out);
ss_mat.C1   = [ss_mat.C1, zeros(ss_mat.m_out, ss_mat.n_dof)];

% Contact force velocity feedback matrix
ss_mat.C2   = [ 1,  0,  h_G;        ...
                1,  0,  h_G;        ...
                0,  1,  -w_core/2;  ...
                0,  1,  w_core/2 ];
ss_mat.C2   = [ zeros(ss_mat.n_u_in, ss_mat.n_dof), ss_mat.C2 ];

%% Natural frequency
wn          = damp(ss_mat.A);   % Structure's natural frequency (rad/sec)
fn          = wn/(2*pi);        % Structure's natural frequency (Hz)
fNyq        = 2*max(fn);        % Nyquist frequency (Hz)

%% Sampling frequency (greater than fNyq)
% fs          = 200;              % Near the Nyquist frequency
fs          = 3200;             % ~40 times the structure's natural frequency
ss_mat.h    = 1/fs;             % Sampling period

%% Noise
% Process noise
ss_mat.Qf   = 3 * ss_mat.h * eye(ss_mat.n_w_in);    % Variance of f (Brownian noise, "random walk"???)
swf         = chol(ss_mat.Qf);                      % Standard deviation of process noise
w           = @(k) (swf*randn(ss_mat.n_w_in, 1));   % Process noise function

% Measurement noise
ss_mat.R    = (1e-2)*eye(ss_mat.m_out);             % Variance of v
sv          = chol(ss_mat.R);                       % Standard deviation of process noise
v           = @(k) (sv*randn(ss_mat.m_out, 1));     % Measurement noise function

%% Code from makemodel.m: Normalization, Change of Basis
% Introducing normalization and change of basis causes the model's matrix
% Sxxc to have elements with magnitudes on the order of 1e-20

%%% Lyapanov Equations (normalizing outputs)
P           = lyap(ss_mat.A, ss_mat.B1 * ss_mat.Qf * ss_mat.B1');    
norm1       = sqrt( trace(ss_mat.C1 * P * ss_mat.C1') );

% Normalize outputs
ss_mat.C1   = ss_mat.C1/norm1;

%% Scale the state-space representation (Change of Basis)
omega_c     = norm(Omega);          % Norm of the matrix of natural frequencies

T           = inv(diag([1,          1,          180/pi, ...
                        1/omega_c,  1/omega_c,  1/omega_c*180/pi])); % Scaling matrix

ss_mat.A    = T \ ss_mat.A * T;     % New basis: z = inv(T)*x, x = T*z
ss_mat.B1   = T \ ss_mat.B1;        % z_{k+1} = inv(T)*x_{k+1} = inv(T) * (A x_k + B u_k)
ss_mat.B2   = T \ ss_mat.B2;        % z_{k+1} = T\A*(T*z_k) + T\B*u_k
ss_mat.C1   = ss_mat.C1 * T;     
ss_mat.C2   = ss_mat.C2 * T;
% y_k = C*x_k = C*T*z_k

%%%%% check the steady state solution
Sigma       = lyap(ss_mat.A, ss_mat.B1 * ss_mat.Qf * ss_mat.B1');	% State error variance matrix

disp('-----')
norm(Sigma)                                        	% State variance
norm(ss_mat.C1 * Sigma * ss_mat.C1' + ss_mat.R)     % Innovation variance

%% True parameters
% For alpha = 0 and beta = 0, y2max = 0.87. Set vc = (1/beta) < 0.87, or
% else the contact force will behave like linear-viscous damping.
% vc      = 1/10;         % 1e-3 < vc < 0.87 for hyperbolic tangent model
% beta    = 1/vc;         % For vc = 0.1, beta = 10
% gamma   = 1e3;          % 10 < gamma < 5e4 for underdamped system with significant contact
% alpha   = gamma*vc;     % For gamma = 1e3 and vc = 0.1, alpha = 100

% % Parameters for contactless simulation:
% alpha   = 0;
% beta    = 0;
% gamma   = alpha*beta;

% Setting parameters directly:
alpha   = 100;
beta    = 10;
gamma   = alpha*beta;

%% Equivalent Damping of Contact Force
disp('Damping Norms')

% Structural Damping
disp('Structure alone:')
damp(ss_mat.A)
norm(C_damp)

% Contact force damping
disp('Structure with Contact:')
A_contact = ss_mat.B2*alpha*beta*ss_mat.C2;
damp(ss_mat.A - A_contact)
norm(A_contact)

disp('Fastest natural frequency w/ contact')
wn_contact = damp(ss_mat.A - A_contact);
fn_contact = max(wn_contact/(2*pi))

%% Allocate space
N_time  = 1e5;
x       = zeros(2*ss_mat.n_dof, N_time);
y       = zeros(ss_mat.m_out, N_time);
time    = 0:ss_mat.h:ss_mat.h*(N_time - 1);

% Iterate
for i   = 2:N_time
    wCT_    = w(i - 1);
    fturb   = ss_mat.B1 * wCT_;
    x(:, i) = myRK4(i - 1, x(:, i - 1), ss_mat, alpha, beta, fturb);
    y(:, i) = ss_mat.C1 * x(:, i) + v(i);
end

% Evaluate contact force
y2  = ss_mat.C2 * x;
fc  = -fcontact(y2, alpha, beta);

%% Plots
t   = 1:N_time;

% Measured time response
figure(1),
subplot(4,1,1)
plot(t, y(1, :));
title('System Outputs')
xlabel('Time');
ylabel('Measurement 1');
subplot(4,1,2)
plot(t, y(2, :));
xlabel('Time');
ylabel('Measurement 2');
subplot(4,1,3)
plot(t, y(3, :));
xlabel('Time');
ylabel('Measurement 3');
subplot(4,1,4)
plot(t, y(4, :));
xlabel('Time');
ylabel('Measurement 4');

% Position
figure(2),
subplot(3,1,1)
plot(t, x(1,:));
xlabel('Time');
ylabel('Horizontal Position');
title('Position States')
subplot(3,1,2)
plot(t, x(2,:));
xlabel('Time');
ylabel('Vertical Position');
subplot(3,1,3)
plot(t, x(3,:));
xlabel('Time');
ylabel('Rotational Position');

% Velocity
figure(3),
subplot(3,1,1)
plot(t, x(4,:));
xlabel('Time');
ylabel('Horizontal Velocity');
title('Velocity States')
subplot(3,1,2)
plot(t, x(5,:));
xlabel('Time');
ylabel('Vertical Velocity');
subplot(3,1,3)
plot(t, x(6,:));
xlabel('Time');
ylabel('Rotational Velocity');

% Phase plots
figure(4),
subplot(3,1,1)
plot(x(1, :), x(4, :))
xlabel('y-position')
ylabel('y-velocity')
title('Phase Plane')
subplot(3,1,2)
plot(x(2,:), x(5,:))
xlabel('z-position')
ylabel('z-velocity')
subplot(3,1,3)
plot(x(3,:), x(6,:))
xlabel('theta-position')
ylabel('theta-velocity')

% Contact model
figure(5)
subplot(4,1,1)
plot(y2(1,:), fc(1,:), '.');
xlabel('Contact velocity 1 (y)')
ylabel('Contact force 1 (y)')
title('Contact force model')
subplot(4,1,2)
plot(y2(2,:), fc(2,:), '.');
xlabel('Contact velocity 2 (y)')
ylabel('Contact force 2 (y)')
subplot(4,1,3)
plot(y2(3,:), fc(3,:), '.');
xlabel('Contact velocity 1 (z)')
ylabel('Contact force 1 (z)')
subplot(4,1,4)
plot(y2(4,:), fc(4,:), '.');
xlabel('Contact velocity 2 (z)')
ylabel('Contact force 2 (z)')

%% Save Data
% Save synthetic data
save('data_core.mat', 'y', 'x');
% Save SS model information (SS matrices, dimensions, Qf, R)
save('model_core.mat', 'ss_mat');
% Save contact info (parameters for grid search)
save('contact_core.mat', 'alpha', 'beta');

% Save CSV file for true response (grid search method)
ytrue_mat = [   0:4;
                time',      y'  ];
writematrix(ytrue_mat, 'Synth_Outputs_True_core.csv')

% Save CSV file for true coefficient combination
coefs = [1; alpha; beta];
writematrix(coefs, 'Synth_Outputs_True_Coefs_core.csv')

% Save CSV file for true states (UKF)
xtrue_mat = [   0:6;
                time',      x'];
writematrix(xtrue_mat, 'Synth_States_True_core.csv')
