% 06/03/2022
% Nicholas Harn
% Grid search
clear;
close all;
clc;
delete(gcp('nocreate'));

% Set rng seed for each replication (shuffle)
rng('shuffle');

% Equivalent to Rep_Gen_CT.m, NOT True_Data_Gen_CT.m

%% Load Data
load('data_core.mat');      % Load synthetic data
load('model_core.mat');     % Load core barrel model
load('contact_core.mat');   % Load contact parameters

%% System dimensions
% Number of states (including parameters)
ss_mat.n_state  = 2*ss_mat.n_dof;
% Number of process noise inputs (including disturbances)
ss_mat.q        = ss_mat.n_w_in;

% Number of augmented states
ss_mat.n_a      = ss_mat.n_state + ss_mat.q + ss_mat.m_out;
% Number of augmented outputs
ss_mat.m_a      = ss_mat.n_state + ss_mat.m_out;

% Number of time points
ss_mat.N        = length(x);                                % Number of samples
time            = 0 : ss_mat.h : ss_mat.h*(ss_mat.N - 1);   % Time vector

%% Define noise
% Process noise
swf = chol(ss_mat.Qf);                      % Process noise standard deviation
w   = @(k) (swf*randn(ss_mat.q, 1));        % Process noise function

% Measurement noise
sv  = chol(ss_mat.R);                       % Measurement noise standard deviation
v   = @(k) (sv*randn(ss_mat.m_out, 1));     % Measurement noise function

%% Set-up grid search
% Parameter values
a       = 1:0.5:5;
ss_mat.alpha_vect   = 10.^a;    % Max contact force
b       = -1:0.5:3;
ss_mat.beta_vect    = 10.^b;    % Inverse of characteristic velocity

% Number of coefficients for the grid
ss_mat.N_coef       = length(a);

%% Parfor loop (for multiple replications)
num_reps            = 20;    % Number of replications for the ensemble

% Begin timer and loop
tic;
parfor rep_id       = 1:num_reps    % Parallel computing on 4 workers
% for rep_id          = 1:num_reps    % Iterative computing on 1 worker
    
    % Broadcast variable ss_mat (unaltered variable from outside the
    % parfor-loop) is sent to each worker at the start of the parfor-loop.
    % If ss_mat is large, sending ss_mat to each worker increases the
    % computation time and required resources ("parallel overhead"?)
    
    % Parfor-loop: 66 sec for 4 reps, 158 sec for 8 reps
    % 4 workers compute the reps in parallel (4 reps per set)
    % Second set takes 92 sec (26 sec longer than the first set)???
    
    % For-loop: 220 sec for 4 reps, 432 sec for 8 reps
    % Roughly 8 sec outside the loop, then 53 sec per loop
    
    %% Initializing arrays for storing data
    % Initialize array for storing states of simulations
    X_all_coefs     = zeros(ss_mat.N_coef, ss_mat.N_coef, ss_mat.n_state, ss_mat.N);
    
    % Initialize arrays for storing inputs and outputs
    Y1_all_coefs    = zeros(ss_mat.N_coef, ss_mat.N_coef, ss_mat.m_out, ss_mat.N);
    Y2_all_coefs    = zeros(ss_mat.N_coef, ss_mat.N_coef, ss_mat.n_u_in, ss_mat.N);
    U1_all_coefs    = zeros(ss_mat.N_coef, ss_mat.N_coef, ss_mat.q, ss_mat.N);
    U2_all_coefs    = zeros(ss_mat.N_coef, ss_mat.N_coef, ss_mat.n_u_in, ss_mat.N);
    
    % Set Initial Condition for State
    X_all_coefs(:, :, :, 1) = zeros(ss_mat.N_coef, ss_mat.N_coef, ss_mat.n_state, 1);
    
    %% Iterative solver
    % Iterate over alpha
    for i_alpha     = 1:ss_mat.N_coef
        alpha       = ss_mat.alpha_vect(i_alpha);
        
        % Iterate over beta
        for i_beta  = 1:ss_mat.N_coef
            beta    = ss_mat.beta_vect(i_beta);
            
            % Keep 1e2 <= gamma = alpha * beta <= 1e4
            if alpha * beta > 1.01e4      % 1.01e4?
                continue
            elseif alpha * beta < 9.9e1   % 9.9e1?
                continue
            end
            
            % Initial conditions for the given parameter combination
            X       = zeros(ss_mat.n_state, ss_mat.N);  % Matrix of state vectors
            X(:,1)  = zeros(ss_mat.n_state, 1);
            
            % Initializing arrays for storing forces and outputs of simulations
            Y1      = zeros(ss_mat.m_out, ss_mat.N);    % Matrix of ex-core measurements
            U1      = zeros(ss_mat.q, ss_mat.N);        % Matrix of turbulent forces
            fturb   = zeros(ss_mat.n_state, ss_mat.N);  % Matrix of process noise terms from turbulence
            
            % Iterate over time
            for i_t = 2:ss_mat.N
                % Input vector
                U1(:, i_t - 1)      = w(i_t - 1);
    
                % Process noise from turbulence
                fturb(:, i_t - 1)   = ss_mat.B1 * U1(:, i_t - 1);
    
                % Solve for the state at the next time point using RK4
                X(:, i_t)           = myRK4(i_t - 1, X(:, i_t - 1), ... 
                    ss_mat, alpha, beta, fturb(:, i_t - 1));
    
                % Output vector
                Y1(:, i_t)          = ss_mat.C1*X(:, i_t) + v(i_t);
            end
            
            % Solve for contact force
            Y2      = ss_mat.C2 * X;
            U2      = fcontact(Y2, alpha, beta);
    
            % Save data for this coefficient combination
            X_all_coefs(i_alpha, i_beta, :, :)  = X;
            Y1_all_coefs(i_alpha, i_beta, :, :) = Y1;
            Y2_all_coefs(i_alpha, i_beta, :, :) = Y2;
            U1_all_coefs(i_alpha, i_beta, :, :) = U1;
            U2_all_coefs(i_alpha, i_beta, :, :) = U2;
        end
    end
    
    %% Save synthetic grid data
    % Define reorganized matrix for storing synthetic data
    output_matrix_1 = zeros(ss_mat.N, ss_mat.N_coef^2 + 1);
    output_matrix_1(:, 1) = time';
    
    i_col = 0;
    % Reorganize the data
    for i_alpha = 1:ss_mat.N_coef
        for i_beta = 1:ss_mat.N_coef
            
            if ss_mat.alpha_vect(i_alpha) * ss_mat.beta_vect(i_beta) > 1.01e4
                continue
            elseif ss_mat.alpha_vect(i_alpha) * ss_mat.beta_vect(i_beta) < 9.9e1
                continue
            end
            
            % Column index
            i_col = i_col + 1;
            
            % Reorganized output data
            out_mat_icol = squeeze(Y1_all_coefs(i_alpha, ...
                i_beta, 1, :));
            output_matrix_1(:, i_col + 1) = out_mat_icol;
        end
    end
    
    % Label the reorganized matrix's columns
    output_matrix_1_labelled = [0:ss_mat.N_coef^2; ...
        output_matrix_1];
    
    % Save matrix as a CSV datafile
    writematrix(output_matrix_1_labelled, sprintf('Synth_Output1_Grid_core_Rep_%03d.csv', rep_id));

end

% Delete parallel pool once finished
delete(gcp('nocreate'));

% End timer
toc

%% Save coefficient data
% Initialize the coefficient matrix
coefs = zeros(2, ss_mat.N_coef^2);

% Populate the coefficient matrix
i_col = 0;
for i_alpha     = 1:ss_mat.N_coef
    for i_beta  = 1:ss_mat.N_coef
        
        if ss_mat.alpha_vect(i_alpha) * ss_mat.beta_vect(i_beta) > 1e4
            continue
        elseif ss_mat.alpha_vect(i_alpha) * ss_mat.beta_vect(i_beta) < 1e2
            continue
        end
        
        % Indexing the data matrix's empty columns (2 - 120)
        i_col   = i_col + 1;

        % Populating each column of the data AND coefficient matrices
        coefs(:, i_col)	= [ss_mat.alpha_vect(i_alpha); ss_mat.beta_vect(i_beta)]; 
    end
end

% Label the coefficient matrix
coef_label      = 1:length(coefs);
coef_matrix_labelled    = [coef_label; coefs];

% Save coefficient data in .csv file
writematrix(coef_matrix_labelled, 'Synth_Output1_Coefs_core.csv');
