% 07/20/2022
% Nicholas Harn
% Unscented Kalman filter for Core Barrel
clear;
close all;
clc;
rng('shuffle');

%% Load Data
load('data_core.mat');      % Load synthetic data
load('model_core.mat');     % Load core barrel model

% Load core contact parameters
load('contact_core.mat');   % Load core contact parameters

%% System dimensions
% Number of states (including parameters)
ss_mat.n_state  = 2*ss_mat.n_dof + 2;
% Number of process noise inputs (including disturbances)
ss_mat.q        = ss_mat.n_w_in + 2;

% Number of augmented states
ss_mat.n_a      = ss_mat.n_state + ss_mat.q + ss_mat.m_out;
% Number of augmented outputs
ss_mat.m_a      = ss_mat.n_state + ss_mat.m_out;

% Number of time points
ss_mat.N        = length(x);                    % Number of samples
time = 0 : ss_mat.h : ss_mat.h*(ss_mat.N - 1);  % Time vector

%% Define noise on the system
% Disturbance of alpha's estimate
seps1       = 1e1;      % Set standard dev of alpha's estimate to 10% of alpha
% seps1       = 0.1*alpha;
Qeps1       = seps1^2;  % Variance of alpha's estimate

% Disturbance of beta's estimate
seps2       = 1e0;      % Set standard dev of beta's estimate to 10% of beta
% seps2       = 0.1*beta;
Qeps2       = seps2^2;  % Variance of beta's estimate

% State var                                                                                                                                                                                                                         iance (with parameter disturbances)
ss_mat.Q    = [ ss_mat.Qf,                  zeros(ss_mat.n_w_in, 2);    ...
                zeros(1, ss_mat.n_w_in),    Qeps1,      0;              ...
                zeros(1, ss_mat.n_w_in),    0,          Qeps2 ];

%% Initialize
% Initial guess of state estimate
% alpha0      = 1e1;  % Underestimate as 10% of the true value 
% beta0       = 1e0;  % Underestimate as 10% of the true value
alpha0      = 0;
beta0       = 0;
xc0         = [ zeros(2*ss_mat.n_dof, 1); alpha0; beta0 ];

% Initial guess of state variance (assuming variance of 1e-3 for ALL states???)
Sxxc        = (1e-3) * eye(ss_mat.n_state);

% Set weighting factor kappa
kappa       = 0.5;

%% Generate the state estimate
[xp, xc, yp, yc, K_t, Sxxc]	= myUKF(y, xc0, Sxxc, kappa, ss_mat);

%% Estimates of alpha and beta
alpha_avg   = mean(xc(7, ss_mat.N/5:ss_mat.N));
alpha_stdev = std(xc(7, ss_mat.N/5:ss_mat.N));
beta_avg    = mean(xc(8, ss_mat.N/5:ss_mat.N));
beta_stdev  = std(xc(8, ss_mat.N/5:ss_mat.N));

%% Create plots of state, output, and parameter estimates
PLOT_SOME_STUFF

%% Save data as CSV
% Create dataset label
data_label = 1:(1 + ss_mat.m_a);

% Organize data (Column 1 = Time, Columns 2 - 7 = States, Columns 8 - 9 =
% Contact Parameters, Columns 10 - 13 = Outputs, Row 1 = Labels, Rows 2 -
% 100001 = Values at various time points)
data_matrix = [ data_label; ...
                time', xc', yc'];

% Save as CSV
writematrix(data_matrix, 'Synth_UKF_Estimates_core.csv')

% Organize data (Column 1 = alpha, Column 2 = beta, Row 1 = Labels, Row 2 =
% Average value, Row 3 = Standard deviation)
param_est   = [ 0,              1;
                alpha_avg,      beta_avg;
                alpha_stdev,    beta_stdev ];

% Save as CSV
writematrix(param_est, 'Synth_UKF_Param_Estimates_core.csv')

% Save outputs as mat file
xc_states = xc(1:2*ss_mat.n_dof, :);
xc_params = xc((2*ss_mat.n_dof + 1):ss_mat.n_state, :);
save('ukf_estimates.mat', 'time', 'xc_states', 'xc_params', 'yc', 'y2c', 'fc_est');

%% FUNCTIONS
% Augmented function for the sigma point transformation
function ybig = fa(t, xbig, ss_mat)
    % Deconstruct the function's augmented input vector xbig
    x_      = xbig(1:ss_mat.n_state, :);                                    % States
    w_      = xbig((ss_mat.n_state + 1):(ss_mat.n_state + ss_mat.q), :);    % Process noise
    v       = xbig((ss_mat.n_state + ss_mat.q + 1):ss_mat.n_a, :);          % Measurement noise
    
    % Deconstruct the state
    xCT_    = x_(1:2*ss_mat.n_dof, :);                      % CT system's states
    param_  = x_((2*ss_mat.n_dof + 1):ss_mat.n_state, :);   % Parameter states
    
    % Deconstruct process noise
    wCT_    = w_(1:ss_mat.n_w_in, :);                       % CT system's process noise
    epsilon = w_((ss_mat.n_w_in + 1):ss_mat.q, :);          % Process noise on parameters
    
    % Compute process noise from turbulence
    fturb   = ss_mat.B1 * wCT_;

    % Update the state
    xCT     = myRK4(t, xCT_, ss_mat, param_(1,:), param_(2,:), fturb);
    
    % Require positive parameters
%     param   = param_ + epsilon;
    param   = abs(param_ + epsilon);
%     param   = abs(0.999*param_ + epsilon);
    
    % Evaluate the measurement (including noise)
    y       = ss_mat.C1 * xCT + v;
    
    % Augmented function's output
    ybig    = [ xCT; param; y ];
end

% UKF prediction and update equation
function [xp, xc, yp, yc, K_t, Sxxc] = myUKF(y, x0, Sxx0, kappa, ss_mat)
    % Allocate space for inputs and outputs
    xp       = zeros(ss_mat.n_state, ss_mat.N);
    xc       = zeros(ss_mat.n_state, ss_mat.N);
    yp       = zeros(ss_mat.m_out, ss_mat.N);
    yc       = zeros(ss_mat.m_out, ss_mat.N);
    
    % Initialize posterior (corrected) estimates
    xc(:, 1) = x0;
    Sxxc     = Sxx0;

    % Create time-varying Kalman gain matrix
    K_t      = zeros(ss_mat.n_state, ss_mat.m_out, ss_mat.N);
    
    % Iterate
    for i    = 2:ss_mat.N
        % Current value of time
        t    = ss_mat.h*i;

        % Assemble augmented state vector
        xbig = [xc(:, i-1); zeros(ss_mat.q + ss_mat.m_out, 1)];
        
        % Assemble augmented state variance
        Xbig = [Sxxc,           zeros(ss_mat.n_state, ss_mat.m_out + ss_mat.q);                     ...
                zeros(ss_mat.q, ss_mat.n_state),    ss_mat.Q,       zeros(ss_mat.q, ss_mat.m_out);  ...
                zeros(ss_mat.m_out, ss_mat.n_state + ss_mat.q),     ss_mat.R];
        
        % Sigma-point transformation
        [ybig, Ybig]    = unscented_trans( @(t, xbig) fa(t, ...
            xbig, ss_mat), t, xbig, Xbig, ss_mat.n_a, ss_mat.m_a, kappa );
        
        % Deconstruct mean
        xp(:, i) = ybig(1:ss_mat.n_state, :);
        yp(:, i) = ybig((ss_mat.n_state + 1) : ss_mat.m_a, :);
        
        % Deconstruct covariance matrix
        Sxxp     = Ybig(1:ss_mat.n_state, 1:ss_mat.n_state);
        Sxyp     = Ybig(1:ss_mat.n_state, (ss_mat.n_state + 1):ss_mat.m_a);
        Syxp     = Ybig((ss_mat.n_state + 1):ss_mat.m_a, 1:ss_mat.n_state);  % Syxp = Sxyp'
        Syyp     = Ybig((ss_mat.n_state + 1):ss_mat.m_a, (ss_mat.n_state + 1):ss_mat.m_a);
        
        % Compute Kalman gain
        K               = Sxyp * inv(Syyp);
        K_t(:, :, i)    = K;
        
        % Correct state estimate
        xc(:, i) = xp(:, i) + K*( y(:, i) - yp(:, i) );
        Sxxc     = Sxxp - Sxyp*inv(Syyp)*Syxp;
        
        % Corrected output
        yc(:, i) = ss_mat.C1*xc(1:2*ss_mat.n_dof, i);
    end
end