function [ybar,Pyy] = unscented_trans(func,t,xbar,Pxx,n,m,kappa)
% UNSCENTED_TRANS The unscented transformation as developed by Julier and
% Uhlmann. Rewritten to account for the time vector t.
%   [ybar,Pyy] = UNSCENTED_TRANS(FUNC,t,XBAR,PXX,M,N,KAPPA) returns the
%   mean and covariance estimate of a function FUNC  for an input
%   probability density with mean XBAR and covariance PXX.  The dimension
%   of the input to FUNC is N and the dimension of the output is M.  KAPPA
%   is a free parameter used in the unscented algorithm.

% Cholesky factorization of the input's variance
U = chol((n+kappa)*Pxx)';   % effectively the input's standard deviation

% initialize matrices for the sigma points, weights, and transformed points
chi = zeros(n, 2*n + 1);
W = zeros(2*n + 1, 1);
y = zeros(m, 2*n + 1);

% transform the mean of the input
chi(:,1) = xbar;
W(1) = kappa/(n + kappa);
y(:,1) = func(t, chi(:,1));

for i = 1:n,
    % sigma points of x (distributed about the mean)
    chi(:,i+1) = xbar + U(:,i);
    chi(:,i+n+1) = xbar - U(:,i);

    % weights of sigma points
    W(i+1) = 1/2/(n + kappa);
    W(i+n+1) = 1/2/(n + kappa);
    
    % transform the sigma points
    y(:,i+1) = func(t,chi(:,i+1));
    y(:,i+n+1) = func(t,chi(:,i+n+1));
end

% initialize the statistics of the transformed points
ybar = zeros(m,1);
Pyy = zeros(m,m);

% evaluate the mean of transformed sigma points
for i = 1:(2*n+1),
    ybar = ybar + W(i)*y(:,i);
end
% evaluate the variance of transformed sigma points
for i = 1:(2*n+1),
    Pyy = Pyy + W(i)*( y(:,i) - ybar )*( y(:,i) - ybar )';
end

return
end
